const coral = @import("coral");

const kym = @import("../kym.zig");

associative: AssociativeTable,
contiguous: ContiguousList,

const AssociativeTable = coral.map.Table(*kym.RuntimeRef, *kym.RuntimeRef, struct {
	pub fn hash(key: *kym.RuntimeRef) usize {
		return key.hash();
	}

	pub fn equals(key_a: *kym.RuntimeRef, key_b: *kym.RuntimeRef) bool {
		return key_a.equals(key_b);
	}
});

const ContiguousList = coral.list.Stack(?*kym.RuntimeRef);

const Self = @This();

pub fn new(env: *kym.RuntimeEnv) kym.RuntimeError!*kym.RuntimeRef {
	var self = Self{
		.associative = AssociativeTable.make(env.allocator, .{}),
		.contiguous = ContiguousList.make(env.allocator),
	};

	errdefer {
		self.associative.free();
		self.contiguous.free();
	}

	const table = try env.new_dynamic(&.{
		.name = "table",
		.size = @sizeOf(Self),
		.destruct = typeinfo_destruct,
		.get = typeinfo_get,
		.set = typeinfo_set,
	});

	coral.io.copy(env.unbox_dynamic(table) catch unreachable, coral.io.bytes_of(&self));

	return table;
}

fn typeinfo_destruct(method: kym.Typeinfo.Method) void {
	const table = @as(*Self, @ptrCast(@alignCast(method.userdata.ptr)));

	{
		var field_iterable = table.associative.as_iterable();

		while (field_iterable.next()) |entry| {
			method.env.discard(entry.key);
			method.env.discard(entry.value);
		}
	}

	table.associative.free();

	while (table.contiguous.pop()) |value| {
		if (value) |ref| {
			method.env.discard(ref);
		}
	}

	table.contiguous.free();
}

fn typeinfo_get(method: kym.Typeinfo.Method, index: *const kym.RuntimeRef) kym.RuntimeError!?*kym.RuntimeRef {
	const table = @as(*Self, @ptrCast(@alignCast(method.userdata.ptr)));
	const acquired_index = try method.env.acquire(index);

	defer method.env.discard(acquired_index);

	if (acquired_index.is_fixed()) |fixed| {
		if (fixed < 0) {
			// TODO: Negative indexing.
			unreachable;
		}

		if (fixed < table.contiguous.values.len) {
			return method.env.acquire(table.contiguous.values[@intCast(fixed)] orelse return null);
		}
	}

	if (table.associative.lookup(acquired_index)) |value_ref| {
		return method.env.acquire(value_ref);
	}

	return null;
}

fn typeinfo_set(method: kym.Typeinfo.Method, index: *const kym.RuntimeRef, value: ?*const kym.RuntimeRef) kym.RuntimeError!void {
	const table = @as(*Self, @ptrCast(@alignCast(method.userdata.ptr)));
	const acquired_index = try method.env.acquire(index);

	errdefer method.env.discard(acquired_index);

	if (acquired_index.is_fixed()) |fixed| {
		if (fixed < 0) {
			// TODO: Negative indexing.
			unreachable;
		}

		if (fixed < table.contiguous.values.len) {
			const maybe_replacing = &table.contiguous.values[@intCast(fixed)];

			if (maybe_replacing.*) |replacing| {
				method.env.discard(replacing);
			}

			maybe_replacing.* = if (value) |ref| try method.env.acquire(ref) else null;

			return;
		}
	}

	const acquired_value = try method.env.acquire(value orelse {
		if (table.associative.remove(acquired_index)) |removed| {
			method.env.discard(removed.key);
			method.env.discard(removed.value);
		}

		return;
	});

	errdefer method.env.discard(acquired_value);

	if (try table.associative.replace(acquired_index, acquired_value)) |replaced| {
		method.env.discard(replaced.key);
		method.env.discard(replaced.value);
	}
}
